--=======================================================================
-- miniSIPServer call waiting service script
--
-- Copyright (C) MYVOIPAPP,Inc. All Rights Reserved.
--
-- author: Gilson
--
-- History:
-- 2013-08-12 Created
-- 2016-08-27 migrated from python script.
--=======================================================================

require "services"
require "utils.dba"
require "utils.date_time"

-- service states
local ECW_IDLE            = 0
local ECW_WAIT_USER_IDLE  = 1
local ECW_RECONN          = 2

-- Timer (Seconds)
local CW_TIMER_MUSIC  = 300 
local CW_TIMER_RECONN = 120

-- audio 
local CW_WAIT_TONE = 0x00080002  -- default music


--====================================
--
-- cache DB functions
--
--====================================
local function cacheDBSaveSession()
    TraceInfo("\t cacheDBSaveSession")

    local createTime = ssfGetCurrSeconds()
    local sqlStr=string.format("insert into tbl_call_waiting (scpFsmID, profileID, createTime)"..
                        " values (%d, %d, %d)", fsm.scpFsmID, fsm.profileID, createTime)
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OK then
        TraceInfo("\t\t fail to insert data. SQL: "..sqlStr)
        return
    end

    fsm.saveSessionInd = MSS_YES                             
end


local function cacheDBDeleteSession()
    TraceInfo("\t cacheDBDeleteSession")
    if fsm.saveSessionInd==MSS_NO then return end

    local sqlStr=string.format("delete from tbl_call_waiting where scpFsmID=%d", fsm.scpFsmID)
    local result = cdbQuery(sqlStr)
    if result~= MSS_ERR_OK then
        TraceInfo("\t\t fail to delete data. SQL: "..sqlStr)
        return
    end

    fsm.saveSessionInd = MSS_NO
end


--====================================
--
-- FSM common functions
--
--====================================
local function initParams()
    fsm.saveSessionInd = MSS_NO
end


local function procException()
    TraceInfo("\t procException")
    cacheDBDeleteSession()
    SendReleaseCall()
    EndService()
end


local function sendRRBEs()
    SendRRBE(CONTROL_LEGID, DP_T_BUSY,         DP_MON_INTERUPT)
    SendRRBE(CONTROL_LEGID, DP_T_NO_ANSWER,    DP_MON_INTERUPT)
    SendRRBE(CONTROL_LEGID, DP_T_ANSWER,       DP_MON_INTERUPT)
end


local function prepareIVR()
    TraceInfo("\t prepareIVR")
    SendCTR(PASSIVE_LEGID)
end


local function playWaitMusic()
    TraceInfo("\t playWaitMusic")
    local paArg = CreatePA()
    paArg.infoSend_annID_annID = CW_WAIT_TONE

    SendPA(PASSIVE_LEGID, paArg)
end

--====================================
--
-- FSM state functions
--
--====================================
local function onIdleProc(currMsg)
    TraceInfo("Call-waiting: onIdleProc")
    
    initParams()
    cacheDBSaveSession()
    sendRRBEs()
    prepareIVR()
    playWaitMusic()

    StartTimer(CW_TIMER_MUSIC)
    EnterState(ECW_WAIT_USER_IDLE)
end


local function onWaitUserIdle(currMsg)
    TraceInfo("onWaitUserIdle")
    if currMsg==ESCP_EVT_UA_USER_IDLE then
        cacheDBDeleteSession()
        SendReconnect()
        StartTimer(CW_TIMER_RECONN)
        EnterState(ECW_RECONN)
    else
        TraceInfo("\t no IDLE message, kill all.")
        procException()
    end 
end

--2014-09-25 If receive other messages except 'answer', just release service.
local function onReconnProc(currMsg)
    TraceInfo("onReconnProc")
    if currMsg==ESCP_EVT_ERB then
        local currEDP = currMessage.edpEvent
        if currEDP==DP_T_ANSWER then
            TraceInfo("\t user answered.")
            SendDFC(PASSIVE_LEGID)
            SendContinue()
            SendTransClose()
            EndService()
            return

        elseif currEDP==DP_T_BUSY or currEDP==DP_T_NO_ANSWER then
            TraceInfo("\t user is still busy")
            -- since data has been deleted from cacheDB, we have to save it back.
            cacheDBSaveSession()
            StartTimer(CW_TIMER_MUSIC)
            EnterState(ECW_WAIT_USER_IDLE)
            return
            
        else
            procException()
            return
        end
    end

    procException()
end

--====================================
--
-- Service main function
--
--====================================
function DestroyProc()
    TraceInfo("DestroyProc")

    cacheDBDeleteSession()
end

function MainProc(currMsg)
    StopTimer()
    
    if fsm.state==ECW_IDLE then
        onIdleProc(currMsg)
    elseif fsm.state==ECW_WAIT_USER_IDLE then
        onWaitUserIdle(currMsg)
    elseif fsm.state==ECW_RECONN then
        onReconnProc(currMsg)        
    else
        TraceInfo("unknown state: "..fsm.state)
    end
end


