--=======================================================================
--miniSIPServer hunting-group service script
--
--Copyright (C) MYVOIPAPP,Inc. All Rights Reserved.
--
--Author: Gilson
--
--History:
--2014-02-24 Migrated from MSS core.
--2016-09-12 Migrated from python script
--2018-01-10 Get HG group infor from cache DB.
--=======================================================================
require "services"
require "utils.data_string"
require "utils.dba"

--HG strategy (defined in core)
local ELH_NULL = 0
local ELH_ROUND_ROBIN = 1
local ELH_LINEAR      = 2

--operator status
local EHG_OPER_CALL_OFFLINE = 0
local EHG_OPER_CALL_IDLE    = 1
local EHG_OPER_CALL_BUSY    = 2

--HG state
local EHG_IDLE = 0
local EHG_WAIT_OP_ANSWER          = 1
local EHG_WAIT_PROMPT_OP_BUSY_END = 2
local EHG_WAIT_OP_IDLE            = 3

--HG audio ID
local HG_ANN_ID_MUSIC     = 0x00080002
local HG_ANN_ID_OPER_BUSY = 0x08080001


--====================================
--
-- Cache DB functions
--
--====================================
local function dbGetHGInfo()
    TraceInfo("\t dbGetHGInfo")
    local sqlStr=string.format("select strategy, noAnswer, queueInd, maxWaitCnt, maxWaitDuration "..
                 "from tbl_hunting_group where groupID=%u",fsm.groupID)
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OBJ_FOUND then return -1 end

    fsm.hgStrategy = cdbGetRowIntVal(0)
    fsm.hgNoAnswer = cdbGetRowIntVal(1)
    fsm.hgCallQueueInd = cdbGetRowIntVal(2)
    fsm.hgMaxWaitCnt = cdbGetRowIntVal(3)
    fsm.hgMaxWaitDuration = cdbGetRowIntVal(4)
    return 1
end


local function dbRemoveFromCallQueue(groupID, scpFsmID)
    if groupID==MSS_INVALID_OBJ_ID or scpFsmID==MSS_INVALID_OBJ_ID then return -1 end

    local sqlStr=string.format("delete from cache_hg_queue where "..
                 "groupID=%d and fsmID=%d", groupID, scpFsmID)
    cdbQuery(sqlStr)
    return 0                 
end


local function cdbHGOperatorGetLoginCnt()
    local sqlStr=string.format("select count(*) from tbl_hunting_group_operator where "..
                 "groupID=%d", fsm.groupID)
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OBJ_FOUND then return -1 end

    result = cdbGetRowIntVal(0)
    return result                 
end


local function cdbHGOperatorGetIdleCnt()
    local sqlStr = string.format("select count(*) from tbl_hunting_group_operator where "..
                   "groupID=%d and callState=%d", fsm.groupID, EHG_OPER_CALL_IDLE)
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OBJ_FOUND then return -1 end

    result = cdbGetRowIntVal(0)
    return result                       
end


local function cdbHGOperatorSetCallState(operator, callState)
    local sqlStr=string.format("update tbl_hunting_group_operator set "..
                "callState=%u where operator='%s'", callState, operator)
    local result=cdbQuery(sqlStr)
    if result~=MSS_ERR_OK then
        TraceInfo("\t\t SQL error: "..sqlStr)
    end                
end


local function cdbHGOperatorSetRoundRobinCnt(operator, cnt)
    local sqlStr=string.format("update tbl_hunting_group_operator set "..
                "roundRobinCnt=%d where operator='%s'", cnt, operator)
    cdbQuery(sqlStr)                
end


local function cdbHGOperatorGetOperatorForRoundRobin()
    local sqlStr=string.format("select operator, roundRobinCnt from tbl_hunting_group_operator where "..
                 "groupID=%u and callState=%u order by roundRobinCnt limit 1", fsm.groupID, EHG_OPER_CALL_IDLE)
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OBJ_FOUND then return MSS_ERR_OBJ_NOT_FOUND end

    fsm.operator=cdbGetRowStrVal(0)
    fsm.currRoundRobin=cdbGetRowIntVal(1)
    fsm.currRoundRobin = fsm.currRoundRobin + 1
    TraceInfo("\t\t operator="..fsm.operator..", roundRobin="..fsm.currRoundRobin)
    return MSS_ERR_OK              
end


local function cdbHGOperatorGetOperatorForLinear()
    local sqlStr=string.format("select operator from tbl_hunting_group_operator where "..
                 "groupID=%u and callState=%u order by loginTime limit 1", fsm.groupID, EHG_OPER_CALL_IDLE)
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OBJ_FOUND then return MSS_ERR_OBJ_NOT_FOUND end

    fsm.operator = cdbGetRowStrVal(0)
    return MSS_ERR_OK                     
end


local function cdbGetCallQueueCnt(groupID)
    local sqlStr=string.format("select count(*) from cache_hg_queue where groupID=%u",groupID)
    local result=cdbQuery(sqlStr)
    if result~=MSS_ERR_OBJ_FOUND then return -1 end

    result = cdbGetRowIntVal(0)
    return result
end


local function cdbAddToCallQueue(groupID)
    local currSeconds=ssfGetCurrSeconds()
    local sqlStr=string.format("insert into cache_hg_queue (groupID, fsmID, loginTime) "..
                 "values(%d, %d, %d)", groupID, fsm.scpFsmID, currSeconds)
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OK then
        TraceInfo("\t\t SQL fail: "..sqlStr)
        return MSS_ERR_FAIL
    else
        return MSS_ERR_OK
    end                 
end


local function dbGetHGGroupID(dialPlan)
    local sqlStr=string.format("select groupID from tbl_hunting_group_detect where "..
                 "dialPlan='%s' and calledNbr='%s'", dialPlan, fsm.calledNbr)
    local result = cdbQuery(sqlStr)
    if result~=MSS_ERR_OBJ_FOUND then return -1 end

    local groupID = cdbGetRowIntVal(0)
    return groupID
end

--====================================
--
-- Core functions
--
--====================================
local function operatorIdle()
    TraceInfo("\t operatorIdle")
    ssfHGOperatorIdle(fsm.operator)
end

--====================================
--
-- FSM common functions
--
--====================================
local function initParams()
    --HG information
    fsm.groupID = 0
    fsm.hgStrategy = ELH_NULL
    fsm.hgNoAnswer = 0
    fsm.hgCallQueueInd = MSS_NO
    fsm.hgMaxWaitCnt = 0
    fsm.hgMaxWaitDuration = 60

    --operator SRF information
    fsm.opSrfID = 0

    --run-time variants
    fsm.failedCnt = 0
    fsm.operator = ""
    fsm.addToQueueInd = MSS_NO
    fsm.currRoundRobin = 0
    fsm.ctrInd = MSS_NO
end

local function startWaitOperatorTimer()
    TraceInfo("\t startWaitOperatorTimer")
    StartTimer(fsm.hgNoAnswer)
end

local function getGroupID()
    TraceInfo("\t getGroupID")
    local dialPlan = fsm.origDialPlan
    local result = dbGetHGGroupID(dialPlan)
    if result<0 then
        if dialPlan~=MSS_DEFAULT_DIAL_PLAN then
            result = dbGetHGGroupID(dialPlan)
            if result<0 then return -2 end
        else
            return -1            
        end
    end

    fsm.groupID = result
    TraceInfo("\t\t group ID="..result)
    return 0
end


local function removeFromCallQueue()
    TraceInfo("\t removeFromCallQueue")
    if fsm.addToQueueInd==MSS_YES then
        dbRemoveFromCallQueue(fsm.groupID, fsm.scpFsmID)
        fsm.addToQueueInd = MSS_NO
    end
end


local function releaseAllSSF()
    TraceInfo("\t releaseAllSSF")
    SendReleaseCall(CAUSE_NORMAL_UNSPECIFIED,0)
    SendReleaseCall(CAUSE_NORMAL_UNSPECIFIED,1)
end


local function endAllSsf()
    TraceInfo("\t endAllSsf")
    SendTransClose(0)
    SendTransClose(1)
end


local function sendCTR()
    if fsm.ctrInd==MSS_YES then return end
    TraceInfo("\t sendCTR")

    fsm.ctrInd = MSS_YES
    SendCTR(PASSIVE_LEGID)
end


local function promptAllOperatorsBusy(queueID)
    TraceInfo("\t promptAllOperatorsBusy")
    sendCTR()

    local paArg = CreatePA()
    paArg.infoSend_annID_annID = HG_ANN_ID_OPER_BUSY
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_varType = ESRF_INFO_VAR_DIGIT
    paArg.infoSend_var_digit = queueID
    SendPA(PASSIVE_LEGID, paArg)
end


local function bridge2Parties()
    TraceInfo("\t bridge2Parties")
    local b2pArg = CreateBridge2Parties()
    b2pArg.peerSrfID = fsm.opSrfID
    b2pArg.peerSsfFsmID = GetSsfFsmID(1)
    SendBridge2Parties(b2pArg)
end


local function saveIDPInfo()
    TraceInfo("\t saveIDPInfo")    
    local result = getGroupID()
    if result<0 then
        TraceInfo("\t\t fail to detect group ID. errCode="..result)
        return -1
    end
    return 0
end


local function exceptionProc()
    TraceInfo("\t exceptionProc")
    releaseAllSSF()
    operatorIdle()
    EndService()
end


local function monitorFirstCall()
    TraceInfo("\t monitorFirstCall")
    SendRRBE(PASSIVE_LEGID, DP_T_DISCONNECT,DP_MON_NOTIFY)
end


local function monitorSecondCall()
    TraceInfo("\t monitorSecondCall")
    SendRRBE(PASSIVE_LEGID, DP_O_ANSWER,DP_MON_INTERUPT,1)
end

local function playMusic()
    TraceInfo("\t playMusic")
    sendCTR()
    SendMusic(PASSIVE_LEGID, HG_ANN_ID_MUSIC)
end


local function discMusic()
    TraceInfo("\t discMusic")
    if fsm.ctrInd==MSS_NO then return end

    SendDFC(PASSIVE_LEGID)
    fsm.ctrInd = MSS_NO
end


local function getOperatorByRoundRobin()
    TraceInfo("\t getOperatorByRoundRobin")
    local result = cdbHGOperatorGetOperatorForRoundRobin()
    if result~=MSS_ERR_OK then return MSS_ERR_OBJ_NOT_FOUND end

    cdbHGOperatorSetRoundRobinCnt(fsm.operator, fsm.currRoundRobin)
    cdbHGOperatorSetCallState(fsm.operator, EHG_OPER_CALL_BUSY)
    return MSS_ERR_OK
end


local function getOperatorByLinear()
    TraceInfo("\t getOperatorByLinear")
    local result = cdbHGOperatorGetOperatorForLinear()
    if result~=MSS_ERR_OK then return MSS_ERR_OBJ_NOT_FOUND end

    cdbHGOperatorSetCallState(fsm.operator, EHG_OPER_CALL_BUSY)
    return MSS_ERR_OK
end


local function getOperator()
    TraceInfo("\t getOperator")
    local loginCnt = cdbHGOperatorGetLoginCnt()
    TraceInfo("\t\t login count = "..loginCnt)
    if loginCnt<=0 then return MSS_ERR_OBJ_NOT_FOUND end

    local idleCnt = cdbHGOperatorGetIdleCnt()
    TraceInfo("\t\t idle count = "..idleCnt)
    if idleCnt<=0 then return MSS_ERR_BUSY end

    if fsm.failedCnt>=loginCnt then
        TraceInfo("\t\t failed so many times, stop trying.")        
        return MSS_ERR_BUSY
    end

    if fsm.hgStrategy==ELH_ROUND_ROBIN then
        getOperatorByRoundRobin()
    elseif fsm.hgStrategy==ELH_LINEAR then 
        getOperatorByLinear()
    else
        TraceInfo("\t\t unknown strategy: "..fsm.hgStrategy)
        return MSS_ERR_INVALID_OPR
    end

    return MSS_ERR_OK
end


local function findOperator()
    TraceInfo("\t findOperator")
    local result = getOperator()
    if result==MSS_ERR_OK then 
        return MSS_ERR_OK 
    elseif result==MSS_ERR_BUSY then
        TraceInfo("\t\t cannot find an idle operator")
        return MSS_ERR_BUSY
    else
        TraceInfo("\t\t no operator.")
        return MSS_ERR_OBJ_NOT_FOUND        
    end
end


local function callOperator()
    TraceInfo("\t callOperator")
    local icaArg=CreateICA()
    icaArg.calledNbr = fsm.operator
    icaArg.callerNbrInd = MSS_YES
    icaArg.callerNbr = fsm.callerNbr
    icaArg.callerSrfInfoInd = MSS_YES
    icaArg.callerSrfID = fsm.callerSrfID

    SendICA(icaArg, 1)
end


local function allOperatorsBusy()
    TraceInfo("\t allOperatorsBusy")
    if fsm.hgCallQueueInd==MSS_NO then
        TraceInfo("\t\t this group can't support queue, so release call.")
        SendReleaseCall()
        EndService()
        return MSS_ERR_INVALID_OPR
    end

    local queueID = cdbGetCallQueueCnt(fsm.groupID)
    if queueID<0 or queueID>fsm.hgMaxWaitCnt then
        TraceInfo("\t\t cannot add call to queue.")
        SendReleaseCall()
        EndService()
        return MSS_ERR_FAIL
    end

    local result = cdbAddToCallQueue(fsm.groupID)
    if result~=MSS_ERR_OK then
        TraceInfo("\t\t fail to add into queue.")
        SendReleaseCall()
        EndService()
        return MSS_ERR_OUT_OF_MEMORY
    end

    fsm.addToQueueInd = MSS_YES
    fsm.failedCnt = 0 -- reset it to prepare next process to find operators
    queueID = queueID + 1    
    promptAllOperatorsBusy(queueID)

    StartTimer(fsm.hgMaxWaitDuration)
    EnterState(EHG_WAIT_PROMPT_OP_BUSY_END)
    return MSS_ERR_OK
end


local function callNextOperator()
    TraceInfo("\t callNextOperator")
    StopTimer()
    fsm.operator=""

    local result = findOperator()
    if result~=MSS_ERR_OK then
        if result==MSS_ERR_BUSY then
            allOperatorsBusy()
        else
            SendReleaseCall()
            EndService()
        end
        return
    end

    playMusic()
    callOperator()

    startWaitOperatorTimer()
    EnterState(EHG_WAIT_OP_ANSWER)
end


local function operatorAnswer()
    TraceInfo("\t operatorAnswer")

    StopTimer()
    discMusic()
    bridge2Parties()
    SendContinue(1)
    endAllSsf()
    EndService()
end

--====================================
--
-- FSM state functions
--
--====================================
local function onIdleProc()
    TraceInfo("onIdleProc")
    initParams()

    local result = saveIDPInfo()
    if result<0 then
        SendReleaseCall()
        EndService()
    end

    result = dbGetHGInfo()
    if result<0 then
        SendReleaseCall()
        EndService()
    end

    monitorFirstCall()
    callNextOperator()
end


local function onWaitOperatorAnswer_recvIcaResp()
    monitorSecondCall()
    SendContinue(1)
end


local function onWaitOperatorAnswer_ERB()
    TraceInfo("\t onWaitOperatorAnswer_ERB")
    if currMessage.edpEvent==DP_O_ANSWER then
        fsm.opSrfID = currMessage.calledSrfID
        operatorAnswer()
    else
        TraceInfo("\t\t first call is released, destroy whole process.")
        exceptionProc()
    end
end


local function onWaitOperatorAnswer_TransClose()
    TraceInfo("\t onWaitOperatorAnswer_TransClose") 
    local ssfFsmID = GetSsfFsmID(0)
    if ssfFsmID==MSS_INVALID_OBJ_ID then
        --if the message is from first call, that means exception
        exceptionProc()
    else
        fsm.failedCnt = fsm.failedCnt + 1
        operatorIdle()
        callNextOperator()
    end 
end


local function onWaitOperatorAnswer_TimeOut()
    TraceInfo("\t onWaitOperatorAnswer_TimeOut")
    fsm.failedCnt = fsm.failedCnt + 1
    SendReleaseCall(CAUSE_NORMAL_UNSPECIFIED, 1)
    operatorIdle()
    callNextOperator()
end


local function onWaitOperatorAnswer(currMsg)
    TraceInfo("onWaitOperatorAnswer")
    if currMsg==ESCP_EVT_ICA_RESP then
        onWaitOperatorAnswer_recvIcaResp()

    elseif currMsg==ESCP_EVT_ERB then
        onWaitOperatorAnswer_ERB()

    elseif currMsg==ESCP_EVT_TRANS_CLOSE then
        onWaitOperatorAnswer_TransClose()

    elseif currMsg==ESCP_EVT_TIMEOUT then
        onWaitOperatorAnswer_TimeOut()

    else
        TraceInfo("\t unknown message: "..currMsg)
    end
end


local function onWaitAnnEnd(currMsg)
    TraceInfo("onWaitAnnEnd")
    if currMsg==ESCP_EVT_SRR then
        StopTimer()
        playMusic()
        StartTimer(fsm.hgMaxWaitDuration)
        EnterState(EHG_WAIT_OP_IDLE)

    elseif currMsg==ESCP_EVT_DB_OPERATOR_IDLE then
        SaveMessage()

    else
        StopTimer()
        exceptionProc()
    end
end


local function onWaitOPIdle(currMsg)
    TraceInfo("onWaitOPIdle")
    StopTimer()

    if currMsg==ESCP_EVT_DB_OPERATOR_IDLE then
        removeFromCallQueue()
        callNextOperator()        
    else        
        exceptionProc()
    end
end

--====================================
--
-- Service main function
--
--====================================
function DestroyProc()
    TraceInfo("DestroyProc")
    removeFromCallQueue()
end


function MainProc(currMsg)
    if fsm.state==EHG_IDLE then
        onIdleProc()

    elseif fsm.state==EHG_WAIT_OP_ANSWER then
        onWaitOperatorAnswer(currMsg)

    elseif fsm.state==EHG_WAIT_PROMPT_OP_BUSY_END then
        onWaitAnnEnd(currMsg)

    elseif fsm.state==EHG_WAIT_OP_IDLE then
        onWaitOPIdle(currMsg)

    else
        TraceInfo("unknown state: "..fsm.state)
        EndService()
    end
end
