--=======================================================================
-- miniSIPServer music oh hold service script
--
-- Copyright (C) MYVOIPAPP,Inc. All Rights Reserved.
--
-- author: Gilson
--
-- History:
-- 2014-02-10 Migrated from MSS core.
-- 2016-08-15 Migrated from Python script.
--=======================================================================
require "services"

-- service states
local EMOH_STATE_IDLE = 0
local EMOH_STATE_PLAY_MUSIC = 1

-- Timer values (seconds)
local MOH_TIMER_MUSIC_DURATION = 1200

--====================================
--
-- IVR functions
--
--====================================
local function stopMusic()
    TraceInfo("\t stopMusic")
    if fsm.ctrInd==MSS_NO then return end

    fsm.ctrInd = MSS_NO
    SendDFC(PASSIVE_LEGID)
end

local function playMusic()
    TraceInfo("\t playMusic")
    SendCTR(PASSIVE_LEGID)
    fsm.ctrInd = MSS_YES

    local paArg = CreatePA()
    paArg.discForbid = AUTO_DISCONNECT_ALLOWED
    paArg.notifyEnd = MSS_YES
    paArg.notifyBegin = MSS_YES
    paArg.infoSend_duration = MOH_TIMER_MUSIC_DURATION
    paArg.infoSend_annID_annID = ANN_ID_COMMON_MUSIC
    SendPA(PASSIVE_LEGID, paArg)
end

local function annEnd()
    TraceInfo("\t annEnd")
    fsm.ctrInd = MSS_NO
    SendContinue()
    SendTransClose()
    EndService()
end
--====================================
--
-- FSM common functions
--
--====================================
local function checkIDPParam()
    TraceInfo("\t checkIDPParam")
    local mediaDir = fsm.mediaDir
    if mediaDir~=ESDP_ATTR_SENDONLY and mediaDir~=ESDP_ATTR_RECVONLY and mediaDir~= ESDP_ATTR_INACTIVE then
        return -1
    end
    return 0
end

local function sendRRBEs()
    TraceInfo("\t sendRRBEs")
    SendRRBE(PASSIVE_LEGID, DP_O_DISCONNECT, DP_MON_NOTIFY)
    SendRRBE(PASSIVE_LEGID, DP_T_DISCONNECT, DP_MON_NOTIFY)
    SendRRBE(CONTROL_LEGID, DP_O_DISCONNECT, DP_MON_NOTIFY)
    SendRRBE(CONTROL_LEGID, DP_T_DISCONNECT, DP_MON_NOTIFY)
    SendRRBE(CONTROL_LEGID, DP_T_MID_CALL, DP_MON_INTERUPT) -- for call-park service
end

local function recvSRR()
    TraceInfo("\t recvSRR")
    local srrType = currMessage.srrType
    if srrType==ESRF_SRR_ANN_BEGIN then
        TraceInfo("\t\t begin music ...")
        SendContinue()

    elseif srrType==ESRF_SRR_ANN_END then
        TraceInfo("\t\t end music.")
        annEnd()

    else
        TraceInfo("\t\t unknown SRR type: "..srrType)
        EndService()
    end
end
--====================================
--
-- FSM state functions
--
--====================================
local function onIdleProc(currMsg)
    TraceInfo("MOH: onIdleProc")
    fsm.ctrInd = MSS_NO

    local result = checkIDPParam()
    if result<0 then
        TraceInfo("\t invalid IDP parameters. result="..result)
        EndService()
        return
    end

    sendRRBEs()
    playMusic()

    StartTimer(MOH_TIMER_MUSIC_DURATION + 10)
    EnterState(EMOH_STATE_PLAY_MUSIC)
end

local function onPlayMusicProc(currMsg)
    TraceInfo("onPlayMusicProc")

    if currMsg==ESCP_EVT_SRR then
        recvSRR()

    elseif currMsg==ESCP_EVT_ERB and currMessage.edpEvent==DP_T_MID_CALL then
        -- for call park service
        annEnd()
    else
        EndService()
    end
end
--====================================
--
-- Service main function
--
--====================================
function DestroyProc()
    TraceInfo("DestroyProc")
    SendContinue()
    SendTransClose()
end

function MainProc(currMsg)    
    if fsm.state==EMOH_STATE_IDLE then
        onIdleProc(currMsg)

    elseif fsm.state==EMOH_STATE_PLAY_MUSIC then
        onPlayMusicProc(currMsg)
    
    else
        TraceInfo("unknown state: "..fsm.state)
        EndService()
    end
end


