--=======================================================================
-- miniSIPServer voice mail service script
--
-- Copyright (C) MYVOIPAPP,Inc. All Rights Reserved.
--
-- author: Gilson
--
-- History:
-- 2014-02-15 Migrated from MSS core.
-- 2016-09-01 Migrated from python script.
--=======================================================================
require "services"
require "utils.data_string"
require "utils.dba"


--Voice mail services
local EVM_TYPE_UNKNOWN    = 0
local EVM_TYPE_RECORD_VM  = 1
local EVM_TYPE_HEAR_VM    = 2
local EVM_TYPE_RECORD_GREETING = 3
local EVM_TYPE_HEAR_GREETING   = 4

--service states
local EVM_STATE_IDLE = 0
local EVM_STATE_RECORD_VM_NOTICE  = 1
local EVM_STATE_RECORD_VM         = 2
local EVM_STATE_HEAR_GREETING     = 3
local EVM_STATE_HEAR_VM_PLAY          = 4
local EVM_STATE_HEAR_VM_WAIT_COMMAND  = 5
local EVM_STATE_HEAR_VM_FINAL_WARNING = 6

--announcement ID
local VM_ANN_ID_WELCOME = 0x02080004
local VM_ANN_ID_BEEP    = 0x01080000 -- just a beep tone
local VM_ANN_ID_NO_VOICEMAIL = 0x02080005
local VM_ANN_ID_WAIT_COMMAND = 0x02080003

--digit maps
local VM_FIRST_DIGIT_TIMEOUT = 20
local VM_INTER_DIGIT_TIMEOUT = 4

--digit maps
local VM_TIMER_VAL_PLAY_ANN = 60
local VM_TIMER_RECORD       = 30 --MSS core can only store 30 seconds data defined by MAX_ANN_RECD_SIZE

--====================================
--
-- FSM common functions
--
--====================================
local function initParams()
    fsm.vmMode = ssfGetSysVarVoiceMailMode()
    fsm.pfxRecordVM = ssfGetSysVarVMRecordPrefix()
    fsm.pfxHearVM = ssfGetSysVarVMHearVM()
    fsm.pfxRecordGreeting = ssfGetSysVarVMRcdGreetingPrefix()
    fsm.pfxHearGreeting = ssfGetSysVarVMHearGreetingPrefix() 

    fsm.vmService = EVM_TYPE_UNKNOWN
    fsm.vmWelcomeAnnID = VM_ANN_ID_WELCOME

    fsm.usrName = ""
    fsm.vmTotal = 0
    fsm.vmUnread = 0
    fsm.vmFileList = {}
    fsm.vmFileIndex = 1 --which vm file is to be played
    fsm.vmFileName = "" --current vm file name

    fsm.beginRecordInd = MSS_NO
end


local function sendVMSMWI()
    TraceInfo("\t sendVMSMWI")
    TraceInfo("\t total="..fsm.vmTotal..", unread="..fsm.vmUnread)
    ssfSendVoiceMailMWI(fsm.scpFsmID, fsm.usrName, fsm.vmTotal, fsm.vmUnread)
end


local function getSubVMDir()
    local vmDir = ssfGetVoiceMailDir()
    local subVmDir = vmDir..fsm.usrName
    return subVmDir
end


local function getAllVMFiles()
    local subVmDir = getSubVMDir()
    fsm.vmFileList = ssfFindFiles(subVmDir,"*.wav")
    if fsm.vmFileList==nil then
        return 0
    else
        return #(fsm.vmFileList)
    end
end


local function getVMUnreadTotal()
    TraceInfo("\t getVMUnreadTotal")
    fsm.vmTotal = getAllVMFiles()
    fsm.vmUnread = fsm.vmTotal
end


local function getCurrVMFileName()
    local fileName = fsm.vmFileList[fsm.vmFileIndex]
    TraceInfo("\t current VM file name is "..fileName)
    fsm.vmFileIndex = fsm.vmFileIndex%fsm.vmTotal+1 -- prepare next vm file
    return fileName
end


local function getCurrVMFullFileName(vmFileName)
    local subVmDir = getSubVMDir()
    local fullName = subVmDir.."/"..vmFileName 
    return fullName
end


local function delCurrVoiceMail()
    TraceInfo("\t delCurrVoiceMail")
    local fullName = getCurrVMFullFileName(fsm.vmFileName)
    os.remove(fullName)

    local vmCnt=#(fsm.vmFileList)
    for currFileIndex=1, vmCnt do
        local currFile = fsm.vmFileList[currFileIndex]
        if currFile~=nil and currFile==fsm.vmFileName then
             table.remove(fsm.vmFileList,currFileIndex)
        end
    end

    fsm.vmTotal = #(fsm.vmFileList)
    if fsm.vmTotal==0 then
        fsm.vmUnread = 0
        TraceInfo("\t\t no VM now.")
        return -1
    end

    fsm.vmFileIndex = fsm.vmFileIndex%fsm.vmTotal+1
    return 0
end


local function detectServiceType()
    TraceInfo("\t detectServiceType")
    local calledNbr = fsm.calledNbr    
    if StrHasPrefix(calledNbr, fsm.pfxRecordVM)==MSS_TRUE then
        fsm.vmService = EVM_TYPE_RECORD_VM
    elseif  StrHasPrefix(calledNbr, fsm.pfxHearVM)==MSS_TRUE then
        fsm.vmService = EVM_TYPE_HEAR_VM
    elseif StrHasPrefix(calledNbr, fsm.pfxRecordGreeting)==MSS_TRUE then
        fsm.vmService = EVM_TYPE_RECORD_GREETING
    elseif StrHasPrefix(calledNbr, fsm.pfxHearGreeting)==MSS_TRUE then
        fsm.vmService = EVM_TYPE_HEAR_GREETING
    else
        TraceInfo("\t\t unknown VM service.")
        return -1
    end
    return 0
end


local function checkIDP()
    TraceInfo("\t checkIDP")
    if fsm.vmService==EVM_TYPE_RECORD_VM then
        if string.len(fsm.redirNbr)==0 then
            TraceInfo("\t\t cannot find redir number.") 
            return -1 
        end
    end
    return 0
end


local function playGreeting()
    TraceInfo("\t playGreeting")
    local paArg = CreatePA()
    paArg.infoSend_audioOwner  = fsm.usrName
    paArg.infoSend_annID_annID = fsm.vmWelcomeAnnID
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_duration = VM_TIMER_VAL_PLAY_ANN + 10

    SendPA(CONTROL_LEGID, paArg)
end


local function playGreetingAndBeep()
    TraceInfo("\t playGreetingAndBeep")
    local paArg = CreatePA()
    paArg.infoSend_audioOwner  = fsm.usrName
    paArg.infoSend_annID_annID = fsm.vmWelcomeAnnID
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_duration = VM_TIMER_VAL_PLAY_ANN + 10

    paArg.infoSend_varType = ESRF_INFO_VAR_MULTI_ID
    paArg.infoSend_var_annID = VM_ANN_ID_BEEP

    SendPA(CONTROL_LEGID, paArg)
end


local function playNoVoiceMail()
    TraceInfo("\t playNoVoiceMail")
    local paArg = CreatePA()
    paArg.infoSend_annID_annID = VM_ANN_ID_NO_VOICEMAIL
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_duration = VM_TIMER_VAL_PLAY_ANN + 10

    SendPA(CONTROL_LEGID, paArg)
end


local function playVoicemail(currVmFileName)
    TraceInfo("\t playVoicemail")
    local vmFullName = getCurrVMFullFileName(currVmFileName)
    TraceInfo("\t\t VM full file name is "..vmFullName)
    local paArg = CreatePA()
    paArg.infoSend_annID_annID = 0
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_duration = VM_TIMER_VAL_PLAY_ANN + 10

    paArg.infoSend_varType = ESRF_INFO_VAR_ANN_FILE
    paArg.infoSend_var_annFileName = vmFullName

    SendPA(CONTROL_LEGID, paArg)
end


local function sendPNCWaitCommand()
    TraceInfo("\t sendPNCWaitCommand")
    local pncArg=CreatePNC()
    pncArg.infoSend_annID_annID = VM_ANN_ID_WAIT_COMMAND
    pncArg.infoSend_numOfRepeat = 1

    pncArg.minNumOfDigits = 1
    pncArg.maxNumOfDigits = 1
    pncArg.firstDigitTimeout = VM_FIRST_DIGIT_TIMEOUT
    pncArg.intervalDigitTimeout = VM_INTER_DIGIT_TIMEOUT

    SendPNCWithArg(CONTROL_LEGID, pncArg)
end


local function sendRecordVM()
    TraceInfo("\t sendRecordVM")
    local recordAnnArg = CreateRecordANN()
    if fsm.vmService==EVM_TYPE_RECORD_VM then
        recordAnnArg.type = ERECORD_ANN_TYPE_VM
    elseif fsm.vmService==EVM_TYPE_RECORD_GREETING then
        recordAnnArg.type = ERECORD_ANN_TYPE_GREETING
    else
        return -1
    end

    recordAnnArg.callerNbr = fsm.callerNbr
    recordAnnArg.subNbr = fsm.usrName
    SendRecordANN(CONTROL_LEGID, recordAnnArg)
    return 0
end
--====================================
--
-- FSM state functions
--
--====================================
local function prepareRecordVoiceMail()
    TraceInfo("\t prepareRecordVoiceMail")
    fsm.usrName = fsm.redirNbr
    TraceInfo("\t\t user name is "..fsm.usrName)

    playGreetingAndBeep()
    fsm.beginRecordInd = MSS_YES

    StartTimer(VM_TIMER_VAL_PLAY_ANN)
    EnterState(EVM_STATE_RECORD_VM_NOTICE)
end


local function beginHearVoiceMail()
    TraceInfo("\t beginHearVoiceMail")
    getVMUnreadTotal()
    if fsm.vmTotal==0 then
        TraceInfo("\t\t no voice mail.")
        playNoVoiceMail()
        StartTimer(VM_TIMER_VAL_PLAY_ANN)
        EnterState(EVM_STATE_HEAR_VM_FINAL_WARNING)
        return
    end

    fsm.vmFileName = getCurrVMFileName()
    playVoicemail(fsm.vmFileName)
    StartTimer(VM_TIMER_VAL_PLAY_ANN)
    EnterState(EVM_STATE_HEAR_VM_PLAY)
end


local function hearGreetingVoice()
    TraceInfo("\t hearGreetingVoice")
    playGreeting()
    
    StartTimer(VM_TIMER_VAL_PLAY_ANN)
    EnterState(EVM_STATE_HEAR_GREETING)    
end


local function recordGreetingVoice()
    TraceInfo("\t recordGreetingVoice")
    playGreetingAndBeep()
    fsm.beginRecordInd = MSS_YES

    StartTimer(VM_TIMER_VAL_PLAY_ANN)
    EnterState(EVM_STATE_RECORD_VM_NOTICE)
end


local function prepareNextVoicemail()   
    TraceInfo("\t prepareNextVoicemail")
    if fsm.vmUnread>0 then
        fsm.vmUnread = fsm.vmUnread - 1
    else
        playNoVoiceMail()
        StartTimer(VM_TIMER_RECORD)
        EnterState(EVM_STATE_HEAR_VM_FINAL_WARNING)
        return
    end

    sendVMSMWI()
    fsm.vmFileName = getCurrVMFileName()
    playVoicemail(fsm.vmFileName)

    StartTimer(VM_TIMER_VAL_PLAY_ANN)
    EnterState(EVM_STATE_HEAR_VM_PLAY)
end


local function recvCommand()
    TraceInfo("\t recvCommand")
    if currMessage.digits=="0" then
        delCurrVoiceMail()
    end
    prepareNextVoicemail()
end


local function recvCommandError()
    TraceInfo("\t recvCommandError")
    if currMessage.srfError~=ESRF_ERR_FIRST_DIGIT_TIMEOUT then
        TraceInfo("\t\t stop service since SRF report error: "..currMessage.srfError)
        EndService()
        return
    end

    prepareNextVoicemail()
end


local function onIdleProc()
    TraceInfo("voiceMail: onIdleProc")
    initParams()
    local result = detectServiceType()
    if result<0 then EndService() return end

    --check IDP parameters
    result = checkIDP()
    if result<0 then EndService() return end

    --get greeting ann id
    fsm.vmWelcomeAnnID = VM_ANN_ID_WELCOME
    fsm.usrName = fsm.callerNbr
    SendCTR(CONTROL_LEGID)

    --process different voice-mail procedures
    if fsm.vmService==EVM_TYPE_RECORD_VM then
        prepareRecordVoiceMail()
    elseif fsm.vmService==EVM_TYPE_HEAR_VM then
        beginHearVoiceMail()
    elseif fsm.vmService==EVM_TYPE_HEAR_GREETING then
        hearGreetingVoice()
    elseif fsm.vmService==EVM_TYPE_RECORD_GREETING then
        recordGreetingVoice()
    else
        EndService()
    end
end 


local function onVMAnnNoticeProc(currMsg)
    TraceInfo("onVMAnnNoticeProc")
    if currMsg~=ESCP_EVT_SRR then
        TraceInfo("\t unnecessary to continue service.")
        EndService()
        return
    end

    if fsm.vmService~=EVM_TYPE_RECORD_VM and fsm.vmService~=EVM_TYPE_RECORD_GREETING then
        TraceInfo("\t cannot record vm for invalid service.")
        EndService()
        return
    end
    sendRecordVM()
    StartTimer(VM_TIMER_RECORD+2)
    EnterState(EVM_STATE_RECORD_VM)
end


local function onRecordVMProc()
    TraceInfo("onRecordVMProc")
    EndService()
end


local function onPlayingVoiceMail(currMsg)
    TraceInfo("onPlayingVoiceMail")
    if currMsg~=ESCP_EVT_SRR then
        TraceInfo("\t unnecessary to continue playing VM.")
        EndService()
        return
    end

    -- current VM has been played, now wait for command
    sendPNCWaitCommand()
    StartTimer(VM_TIMER_VAL_PLAY_ANN)
    EnterState(EVM_STATE_HEAR_VM_WAIT_COMMAND)
end


local function onWaitCommand(currMsg)
    TraceInfo("onWaitCommand")
    StopTimer()
    if currMsg==ESCP_EVT_PNC_RESULT then
        recvCommand()
    elseif currMsg==ESCP_EVT_TIMEOUT then
        prepareNextVoicemail()
    elseif currMsg==ESCP_EVT_SRF_ERR then
        recvCommandError()
    else
        TraceInfo("\t unnecessary message. stop service.")
        EndService()
    end
end


--====================================
--
-- Service main function
--
--====================================
function DestroyProc()
    TraceInfo("DestroyProc")
    SendReleaseCall()

    if fsm.vmService==EVM_TYPE_RECORD_VM then
        if fsm.beginRecordInd==MSS_NO then return end
        
        if fsm.vmMode==0 then --send vm through email
            -- SSF will send VM through email
        else
            TraceInfo("\t prepare to send MWI.")
            getVMUnreadTotal()
            sendVMSMWI()
        end
    end
end


function MainProc(currMsg)
    if fsm.state==EVM_STATE_IDLE then
        onIdleProc()
    elseif fsm.state==EVM_STATE_HEAR_GREETING then
        EndService()
    elseif fsm.state==EVM_STATE_RECORD_VM_NOTICE then
        onVMAnnNoticeProc(currMsg)
    elseif fsm.state==EVM_STATE_RECORD_VM then
        onRecordVMProc()
    elseif fsm.state==EVM_STATE_HEAR_VM_PLAY then
        onPlayingVoiceMail(currMsg)
    elseif fsm.state==EVM_STATE_HEAR_VM_WAIT_COMMAND then
        onWaitCommand(currMsg)
    else
        TraceInfo("unknown state: "..fsm.state)
        EndService()
    end
end
